#ifndef PROTO
#define PROTO 1
#include "stdio.h"

/* prototypes of all routines for vogle */

/* source file arcs.c */
void arcprecision( int );
void circleprecision( int );
void arc( float, float, float, float, float );
void sector( float, float, float, float, float );
void circle( float, float, float );

/* source file attr.c */
void  pushattributes(void);
void popattributes(void);
#ifdef DEBUG
  void printattribs( char * );
#endif

/* source file aspect.c */
float getaspect(void);
void getdisplaysize( float *, float * );
void getfactors( float *, float * );
void expandviewport(void);
void unexpandviewport(void);

/* source file buffer.c */
int backbuffer(void);
void frontbuffer(void);
int swapbuffers(void);

/* source file pref.c */
void prefposition( int, int );
void prefsize( int, int );
void getprefposandsize( int *, int *, int *, int * );

/* source file clip.c */
int MakeEdgeCoords( int, Vector );
void clip( register Vector, register Vector );
void quickclip( register Vector, register Vector );
void clipping( int );

/* source file curves.c */
void curvebasis( Matrix );
void curveprecision( int );
void rcurve( Matrix );
void curve( float [4][3] );
void drcurve( int, Matrix );
void curven( int, float [][3] );

/* source file draw.c */
void setdash( float );
void linestyle( unsigned char * );
void dashline( register Vector, register Vector );
void draw( float, float, float );
void draw2( float, float );
void rdraw( float, float, float );
void rdraw2( float, float );
void sdraw2( float, float );
void rsdraw2( float, float );

/* source file drivers.c */
FILE *_voutfile(void);
void voutput( char * );
void verror( char * );   
void vinit( char * );
void pushdev(char *);
void popdev(void);
void vnewdev( char * );
char *vgetdev( char * );
int getkey(void);
int checkkey(void);
int locator( float *, float * );
int slocator( float *, float * );
void clear(void);
void vexit(void);
void color( int );
void mapcolor( int, short, short, short );
int getdepth(void);
void vsetflush( int );
void vflush(void);
void linewidth(int);

/* source file valloc.c */
char *vallocate( unsigned );

/* source file viewing.c */
void polarview( float, float, float, float );
void up( float, float, float );
void lookat( float, float, float, float, float, float, float );
void perspective( float, float, float, float );
void window( float, float, float, float, float, float );
void ortho( float, float, float, float, float, float );
void ortho2( float, float, float, float );

/* source file viewp.c */
void pushviewport(void);
void popviewport(void);
void viewport( float, float, float, float );
void getviewport( float *, float *, float *, float * );

/* source file mapping.c */
void VtoWxy( float, float, float *, float * );
void CalcW2Vcoeffs(void);
int WtoVx( float [] );
int WtoVy( float [] );

/* source file matrix.c */
void copyvector( register Vector, register Vector );
void copymatrix( register Matrix, register Matrix );
void copytranspose( register Matrix, register Matrix );
void getmatrix( Matrix );
void pushmatrix(void);
void popmatrix(void);
void loadmatrix( Matrix );
void mult4x4( register Matrix, register Matrix, register Matrix );
void multmatrix( Matrix );
void identmatrix( Matrix );
void multvector( register Vector, register Vector, register Matrix );
void premultvector( Vector, Vector, Matrix );
#ifdef DEBUG
  void printmat( char *, Matrix );
  void printvect( char *, Vector );
#endif

/* source file move.c */
void move( float, float, float );
void move2( float, float );
void rmove( float, float, float );
void rmove2( float, float );
void smove2( float, float );
void rsmove2( float, float );

/* source file objects.c */
void makeobj( int );
void closeobj(void);
void delobj( int );
int genobj(void);
int getopenobj(void);
void loadobj( int, char * );
void saveobj( int, char * );
void callobj( int );
int isobj( int );

/* source file patches.c */
void patchbasis( Matrix, Matrix );
void patchprecision( int, int );
void patchcurves( int, int );
void patch( Matrix, Matrix, Matrix );
void rpatch( Matrix, Matrix, Matrix, Matrix );
void transformtensor( Tensor, Matrix );
void drpatch( Tensor, int, int, int, int, int, int );

/* source file points.c */
void point( float, float, float );
void point2( float, float );
void spoint2( float, float );

/* source file polygons.c */
void polyfill( int );
void polyhatch( int );
void hatchang( float );
void hatchpitch( float );
void backfacedir( int );
void backface( int );
void polyobj( int, Token [] );
void poly2( int, float [][2] );
void poly( int, float [][3] );
void pmove( float, float, float );
void pdraw( float, float, float );
void makepoly(void);
void closepoly(void);

/* source file rect.c */
void rect( float, float, float, float );
void srect( float, float, float, float );

/* source file tensor.c */
void premulttensor( Tensor, Matrix, Tensor );
void multtensor( Tensor, Matrix, Tensor );
void copytensor( Tensor, Tensor );
void copytensortrans( Tensor, Tensor );

/* source file text.c */
void font( char * );
int numchars(void);
int hershfont( char * );
void getcharsize( char, float *, float * );
void drawchar( int );
void textsize( float, float );
float getfontwidth(void);
float getfontheight(void);
void getfontsize( float *, float * );
void drawstr( char * );
void sdrawstr( float, float, char * );
float strlength( char * );
void boxtext( float, float, float, float, char * );
void boxfit( float, float, int );
void centertext( int );
void textjustify( unsigned );
void xcentertext(void);
void ycentertext(void);
void leftjustify(void);
void rightjustify(void);
void topjustify(void);
void bottomjustify(void);
void fixedwidth( int );
void textang( float );

/* source file newtokens.c */
Token *newtokens( int );

/* source file trans.c */
void translate( float, float, float );
void rotate( float, char );

/* source file scale.c */
void scale( float, float, float );

/* sourc file getgp.c */
void getgp( float *, float *, float * );
void getgp2( float *, float * );
void getgpt( float *, float *, float *, float * );
void sgetgp2( float *, float * );

/* source file yobbaray.c */
void yobbarays( int );

/* source file getstring.c */
int getstring( int, char * );

/* source file /drivers/ps.c */
int _CPS_devcpy(void);
int _PS_devcpy(void);
int _PSP_devcpy(void);
int _PCPS_devcpy(void);

int _LASER_devcpy(void);

/* source file drivers/hpdxy.c */

void _HPGL_A2_devcpy(void);
void _HPGL_A3_devcpy(void);
void _HPGL_A4_devcpy(void);
void _HPGL_A1_devcpy(void);
void _DXY_devcpy(void);

/* source file drivers/tek.c */
int TEK_init(void);
int TEK_exit(void);
int TEK_draw(int x, int y);
int TEK_getkey(void);
int TEK_locator(int *x, int *y);
int TEK_clear(void);
int TEK_font(char *font);
int TEK_char(char c);
int TEK_string(char *s);
int TEK_fill(int n, int x[], int y[]);
void _TEK_devcpy(void);

/* source file drivers/X11.c */
void _X11_devcpy(void);

/* source file drivers/unixplot.c */
void _unixplot_devcpy(void);

/* source file drivers/apollo.c */
void _APOLLO_devcpy(void);

/* source file drivers/sun.c */
void vo_sunview_canvas_size(int sw, int sh);
void _SUN_devcpy(void);

/* source file drivers/hpgt.c */
void _HPGT_devcpy(void);

/* source file drivers/decX11.c */
void _DECX11_devcpy(void);

/* source file drivers/grx.c */
int _grx_devcpy(void);

/* source file drivers/NeXT.c */
void _NeXT_devcpy(void);

/* source file drivers/ibmpc/vga.c */
int vga_init(void);
int vga_exit(void);
void _vga_devcpy(void);


/* source file drivers/ibmpc/vega_col.c */

void vega_color( int );
void vega_setpal(void);    

/* source file drivers/ibmpc/vega_dra.c */

void vega_draw( register int, register int );

/* source file drivers/ibmpc/vega_tex.c */

void vega_string( char * );
void vega_char( char );
void vegachar( int, int, int, int, int );
int vega_font( char * );

/* source file drivers/ibmpc/ega.c */

int ega_init(void);
int ega_exit(void);
void _ega_devcpy(void);

/* source file drivers/ibmpc/pc_text.c */

void pc_string( char * );
int pc_font( char * );

/* source file drivers/ibmpc/pc_locat.c */

void pc_locinit( int, int );     
int pc_locator( int *, int * ); 

/* source file drivers/ibmpc/pc_fill.c */

void pc_fill( int, int *, int * );  
int fill_tri( int [], int [] );

/* source file drivers/ibmpc/pc_getke.c */

int pc_getkey(void);         
int pc_checkkey(void);    

/* source file drivers/ibmpc/setmode.c    */

int setmode( int );  

/* source file drivers/ibmpc/watcom.c    */

void egaschar( int, int, int, int ); 
void egalchar( int, int, int, int );
int ismouse( int, int );
void showmouse(void);
int readmouse( int*,int* );
void vega_clear(void); 
void vega_backbuf(void);  
void vega_frontbuf(void); 
void vega_swapbuf(void); 
void egaline( int, int, int, int, int );
int zsetup(void);

/* source file \drivers\ibmpc\mswin.c */

#endif
